﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ServiceModel;
using System.IO;
using System.Workflow.ComponentModel.Compiler;
using System.ServiceModel.Channels;
using System.ServiceModel.Description;
using System.Reflection;
using System.Workflow.Runtime;
using System.Workflow.Runtime.Hosting;
using Contracts;
using System.Workflow.ComponentModel;
using System.Workflow.Activities;

namespace HostApplication
{
    public class LocalWorkflowServiceHost<TWorkflow,THostService> : WorkflowServiceHost
    {
        #region Constructors 

        static LocalWorkflowServiceHost()
        {
            LocalBinding = new NetNamedPipeContextBinding(NetNamedPipeSecurityMode.None);
            workflowBaseAddress = "net.pipe://localhost/WFLocal/Workflow/";
            hostServiceAddress = "net.pipe://localhost/WFLocal/Host/";

        }
        public LocalWorkflowServiceHost() : base(typeof(TWorkflow))
        {
            this.workflowServiceType = typeof(TWorkflow);
            this.hostServiceType = typeof(THostService);
        }

        #endregion Constructors


        #region Private Fields

        Type workflowServiceType;
        Type hostServiceType;
        ServiceHost hostService;
        static Binding LocalBinding;
        static string workflowBaseAddress;
        static string hostServiceAddress;
        Dictionary<Type, ChannelFactory> factories;
        Dictionary<Type, IClientChannel> proxies;

        #endregion Private Fields

        #region Create endpoints and host service

        protected override ServiceDescription CreateDescription(out IDictionary<string, ContractDescription> implementedContracts)
        {
            ServiceDescription desc =  base.CreateDescription(out implementedContracts);

            //add endpoints for the workflow
            foreach (var contract in implementedContracts)
            {
                desc.Endpoints.Add(new ServiceEndpoint(contract.Value,
                        LocalBinding, GetEndpointAddress(contract.Value.ContractType)));
            }

            return desc;
        }
        

        private EndpointAddress GetEndpointAddress(Type contractType)
        {
            return new EndpointAddress(
                String.Concat(workflowBaseAddress, contractType.FullName));
        }

        IEnumerable<Type> GetContractTypes(Type type)
        {
            List<Type> contractTypes = new List<Type>();

            if (IsServiceContractType(type))
            {
                contractTypes.Add(type);
            }
            contractTypes.AddRange(
                type.FindInterfaces((t,o)=>
                    IsServiceContractType(t), null));

            return contractTypes;
        }

        private bool IsServiceContractType(Type type)
        {
            if (type == null)
                return false;

            object[] sContractAtts = type.GetCustomAttributes(typeof(ServiceContractAttribute), false);
            return sContractAtts != null && sContractAtts.Length > 0;
        }

       
        protected override void OnOpened()
        {
            base.OnOpened();
            hostService = new ServiceHost(hostServiceType);
            
            foreach (Type contractType in GetContractTypes(hostServiceType))
            {
                hostService.AddServiceEndpoint(contractType,
                    LocalBinding, GetHostEndpointAddress(contractType));
            }

            hostService.Open();

            //add client endpoints for the service
            AddClientEndpointsToServiceChannelManager();
        }

        private void AddClientEndpointsToServiceChannelManager()
        {
            List<ServiceEndpoint> endpoints = 
                new  List<ServiceEndpoint>();
                    
            foreach (Type contractType in GetContractTypes(hostServiceType))
            {

                ServiceEndpoint endpoint = new ServiceEndpoint
                    (
                        ContractDescription.GetContract(contractType),
                        LocalBinding,
                        new EndpointAddress(GetHostEndpointAddress(contractType))
                    );
                endpoint.Name = "HostEndpoint-" + contractType.FullName;
                endpoints.Add(endpoint);
            }

            //by using the channel manager, no app.config is required
            WorkflowRuntime runtime = this.Description.Behaviors.Find<WorkflowRuntimeBehavior>().WorkflowRuntime;
            ChannelManagerService chanMan =
                new ChannelManagerService(endpoints);

            runtime.AddService(chanMan);
        }

        private string GetHostEndpointAddress(Type contractType)
        {
            return String.Concat(hostServiceAddress, contractType.FullName);
        }

        #endregion Open ServiceHost for host service

        protected override void OnClose(TimeSpan timeout)
        {
            base.OnClose(timeout);
            if (hostService != null)
                if (hostService.State == CommunicationState.Opened)
                    hostService.Close(timeout);
                else
                    hostService.Abort();
        }

        #region Helper Methods for consumers

        public THostService HostServiceInstance { get; private set; }

        public TServiceContract GetProxy<TServiceContract>()
        {
            ChannelFactory<TServiceContract> factory =  GetChannelFactory<TServiceContract>();

            TServiceContract proxy = factory.CreateChannel();
            ((IClientChannel)proxy).Open();
            return proxy;       
        }

        public string GetWorkflowInstanceID(IClientChannel proxy)
        {
            IContextManager mgr = ((IClientChannel)proxy).GetProperty<IContextManager>();

            return mgr.GetContext()["instanceId"];
        }

        private ChannelFactory<TServiceContract> GetChannelFactory<TServiceContract>()
        {
            ChannelFactory<TServiceContract> factory = null;

            if (factories == null)
                factories = new Dictionary<Type, ChannelFactory>();

            if (!factories.ContainsKey(typeof(TServiceContract)))
            {
                factory = new ChannelFactory<TServiceContract>(LocalBinding, GetEndpointAddress(typeof(TServiceContract)));
                factory.Open();
                factories.Add(typeof(TServiceContract), factory);
            }
            else
            {
                factory = (ChannelFactory<TServiceContract>)factories[typeof(TServiceContract)];
            }
            return factory;
        }

        public TServiceContract GetProxy<TServiceContract>(Guid instanceID)
        {
            if (proxies == null)
                proxies = new Dictionary<Type, IClientChannel>();

            if(proxies.ContainsKey(typeof(TServiceContract)))
            {
                return (TServiceContract)proxies[typeof(TServiceContract)];
            }
            else
            {
                ChannelFactory<TServiceContract> factory =  GetChannelFactory<TServiceContract>();
                TServiceContract proxy = factory.CreateChannel();
                
                IContextManager mgr = ((IClientChannel)proxy).GetProperty<IContextManager>();
                mgr.Enabled = false;
                ((IClientChannel)proxy).Open();

                proxies.Add(typeof(TServiceContract), (IClientChannel)proxy);

                return proxy;
            }
        }

        public void PerformClientOperationWithContext(Guid workflowID, IClientChannel proxy, Action clientOperation)
        {
            //setup the context property for the instance id
            using (OperationContextScope scope = new OperationContextScope((IContextChannel)proxy))
            {
                ContextMessageProperty property =
                    new ContextMessageProperty(
                        new Dictionary<string, string>
                        {
                            {"instanceId", workflowID.ToString()}
                        });
                OperationContext.Current.OutgoingMessageProperties.Add("ContextMessageProperty", property);

                //invoke the client operation for the caller
                clientOperation();
            }
        }

        #endregion Helper methods for consumers
    }
}
